# Greese.pm - Generic Report Engine by Extended Select Expression
# 2002-4 <nakajima@netstock.co.jp>
package Greese;
use Safe;
use strict;
use vars qw($VERSION $NOQUOTENAME $DEFAULTCOLUMN);

$VERSION = 0.92;

# if true, Greese::Report::quotename() has no effect
$NOQUOTENAME = 0;

# used if fields default
$DEFAULTCOLUMN = '1';

# for Perl 5.6- : not use Unicode character string
BEGIN {
	eval { require bytes; };
	bytes->import unless $@;
}

my $Safe = new Safe;

#--------------------------------------------------------------------
# tree classes
#--------------------------------------------------------------------
package Greese::Node;

sub new {
	my($class, $parent, @initargs) = @_;
	my $self = bless {parent => $parent}, ref($class) || $class;
	$self->init(@initargs);
	$parent->add_child($self) if $parent;
	$self;
}

sub new_replace {
	my($class, $parent, $idx, @initargs) = @_;
	my $self = bless {parent => $parent}, ref($class) || $class;
	$self->init(@initargs);
	$parent->set_child($idx, $self) if $parent;
	$self;
}

sub new_instance {
	my($class, $parent, @initargs) = @_;
	my $self = bless {parent => $parent}, ref($class) || $class;
	$self->init(@initargs);
	$parent->add_instance($self) if $parent;
	$self;
}

sub parent {
	my($self) = @_;
	$self->{parent};
}

sub set_parent {
	my($self, $parent) = @_;
	$self->{parent} = $parent;
}

sub clone { 
	my($self) = @_;
	my %clone = %$self;
	my $clone = bless \%clone, ref($self);
	# deep copy
	if( $clone->{children} ) {
		$clone->{children} = [map {$_->clone} @{$clone->{children}}];
	}
	if( $clone->{else} ) {
		$clone->{else} = [map {$_->clone} @{$clone->{else}}];
	}
	$clone;
}

sub dump {
	my($self, $indent) = @_;
	my $type = ref($self);
	$type =~ s/.*:://;
	my $values = exists $self->{value} ? 
		"'".$self->value."'" :
		"{".join(', ', map {ref($self->{$_}) eq 'ARRAY' ? 
		"$_ => [".join(',', @{$self->{$_}})."]" :
		"$_ => '$self->{$_}'"} grep 
		!/^(parent|children)$/, keys %$self)."}";
	$values =~ s/\n/\\n/g;
	print "$indent$type $values\n";
}

sub expvalue {
	my($self, $exp, $param) = @_;
	my $value;
	if( $exp =~ /^param\((.+)\)$/ ) {
		$value = $param->{$1};
	} elsif( $exp =~ /^env\((.+)\)$/ ) {
		$value = $ENV{$1};
	} elsif( $exp =~ /^reval\((.+)\)$/ ) {
		my $code = $1;
		my $result = $Safe->reval($code);
		die "error '$@' in reval($code)" if $@;
		$value = $result;
	} elsif( $exp =~ /^count\((.+)\)$/ ) {
		$value = $self->parent->pcount($1);
	} elsif( $exp =~ /^count\(\)$/ ) {
		$value = $self->parent->tuples;
	} elsif( $exp =~ /^page\(\)/ ) {
		$value = $param->{_page};
	} elsif( $exp =~ /^prev\(\)/ ) {
		$value = $param->{_page} > 1 ? $param->{_page} - 1 : 1;
	} elsif( $exp =~ /^next\(\)/ ) {
		$value = $param->{_page} < $param->{_pages} ? 
			$param->{_page} + 1 : $param->{_pages};
	} elsif( $exp =~ /^pages\(\)/ ) {
		$value = $param->{_pages};
	} else {
		die "unknown expression '$exp'\n";
	}
	$value;
}

#--------------------------------------------------------------------
package Greese::Item;
use vars qw(@ISA);
@ISA = qw(Greese::Node);

sub init {
	my($self, $value) = @_;
	$self->{value} = $value;
}

sub value {
	my($self) = @_;
	$self->{value};
}

#--------------------------------------------------------------------
package Greese::Container;
use vars qw(@ISA);
@ISA = qw(Greese::Node);

sub add_child {
	my($self, $child) = @_;
	push @{$self->{children}}, $child;
}

sub set_child {
	my($self, $idx, $child) = @_;
	$self->{children}[$idx] = $child;
}

sub children {
	my($self) = @_;
	$self->{children} ? @{$self->{children}} : ();
}

sub add_instance {
	my($self, $instance) = @_;
	push @{$self->{instances}}, $instance;
}

sub clr_instance {
	my($self) = @_;
	delete $self->{instances};
}

sub instances {
	my($self) = @_;
	$self->{instances} ? @{$self->{instances}} : ();
}

sub dump {
	my($self, $indent) = @_;
	$self->SUPER::dump($indent);
	if( $self->{children} ) {
		for my $child(@{$self->{children}}) {
			$child->dump("$indent  ");
		}
	}
	if( $self->{else} ) {
		print "${indent}Else\n";
		for my $child(@{$self->{else}}) {
			$child->dump("$indent  ");
		}
	}
}

#--------------------------------------------------------------------
# template classes
#--------------------------------------------------------------------
package Greese::Template;
use FileHandle;

sub esc_char { '\\' }
sub start_tag { '*[' }
sub end_tag { ']*' }
sub escape { $_[1] }
sub postfilter { $_[1] }

sub new {
	my($class, $file) = @_;
	$class = ref($class) if ref($class);
	my $self = bless {file => $file}, $class;
	$self->{filenode} = 
		Greese::TemplateContainer::File->new(undef, $self, $file);
	$self->filenode->parse;
	$self;
}

sub file {
	my($self) = @_;
	$self->{file};
}

sub filenode {
	my($self) = @_;
	$self->{filenode};
}

sub dump {
	my($self, $indent) = @_;
	$self->{filenode}->dump($indent);
}

#--------------------------------------------------------------------
package Greese::Template::Xml;
use vars qw(@ISA);
@ISA = qw(Greese::Template);

sub escape { 
	my($self, $str) = @_;
	$str =~ s/&/&amp;/g;
	$str =~ s/</&lt;/g;
	$str =~ s/>/&gt;/g;
	$str;
}

sub filter_urlescape { 
	my($self, $str) = @_;
	$str =~ s/([^\/:+\$,A-Za-z0-9\-_.!~*'()])/sprintf("%%%02X",ord($1))/ge;
	$str;
}

sub filter_urlunescape { 
	my($self, $str) = @_;
	$str =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
	$str;
}

#--------------------------------------------------------------------
package Greese::Template::Html;
use vars qw(@ISA);
@ISA = qw(Greese::Template::Xml);

sub postfilter { 
	my($self, $str) = @_;
	$str =~ s/<tr>\s*<\/tr>//ig;
	$str;
}

sub filter_nl2br { 
	my($self, $str) = @_;
	$str =~ s/\r?\n/<BR>/g;
	$str;
}

#--------------------------------------------------------------------
package Greese::Template::LaTeX;
use vars qw(@ISA);
@ISA = qw(Greese::Template);

sub escape { 
	my($self, $str) = @_;
	$str =~ s/[#\$%&_\{\}<>\\\^\|~]/\\verb!$&!/g;
	$str;
}

#--------------------------------------------------------------------
package Greese::TemplateItem::Literal;
use vars qw(@ISA);
@ISA = qw(Greese::Item);

sub init {
	my($self, $label, $value) = @_;
	$self->{label} = $label;
	$self->{value} = $value;
}

sub add_string {
	my($self, $string) = @_;
	$self->{value} .= $string;
}

sub make_report {
	my($self, $parent) = @_;
	Greese::ReportItem::Literal->new($parent, $self);
}

#--------------------------------------------------------------------
package Greese::TemplateItem::Field;
use vars qw(@ISA);
@ISA = qw(Greese::Item);

sub init {
	my($self, $label, $value) = @_;
	$self->{label} = $label;
	$self->{value} = $value;
	if( $value !~ /^\'/ ) {
		$self->parent->add_innerfield($value);
		$self->parent->add_usefield($value);
	}
}

sub make_report {
	my($self, $parent) = @_;
	Greese::ReportItem::Field->new($parent, $self);
}

#--------------------------------------------------------------------
package Greese::TemplateItem::Expression;
use vars qw(@ISA);
@ISA = qw(Greese::Item);

sub init {
	my($self, $label, $value) = @_;
	$self->{label} = $label;
	$self->{value} = $value;
	if( $value =~ /^(sum|max|min)\((.+)\)/ ) {
		$self->parent->add_usefield($2);
	} elsif( $value =~ /^r?eval\((.+)\)/ ) {
		my $code = $1;
		while( $code =~ /%\[([^\(\)]+?)\]%/g ) {
			$self->parent->add_usefield($1);
		}
	}
}

sub make_report {
	my($self, $parent) = @_;
	Greese::ReportItem::Expression->new($parent, $self);
}

#--------------------------------------------------------------------
package Greese::TemplateItem::Embed;
use vars qw(@ISA);
@ISA = qw(Greese::Item);

sub init {
	my($self, $args) = @_;
	my($subtemplate, %args) = @$args;
	$self->{subtemplate} = $subtemplate;
	$self->{args} = \%args;
	for my $value(values %args) {
		while( $value =~ /%\[([^\(\)]+?)\]%/g ) {
			my $fname = $1;
			$self->parent->add_innerfield($fname);
			$self->parent->add_usefield($fname);
		}
	}
}

sub make_report {
	my($self, $parent) = @_;
	Greese::ReportItem::Embed->new($parent, $self);
}

#--------------------------------------------------------------------
package Greese::TemplateContainer;
use vars qw(@ISA);
@ISA = qw(Greese::Container);

sub add_usefield {
	my($self, @fields) = @_;
	$self->parent->add_usefield(@fields) if $self->parent;
}

sub add_innerfield {
	my($self, @fields) = @_;
	$self->parent->add_innerfield(@fields) if $self->parent;
}

#--------------------------------------------------------------------
package Greese::TemplateContainer::Repeat;
use vars qw(@ISA);
@ISA = qw(Greese::TemplateContainer);

sub init {
	my($self, $name, $perpage) = @_;
	$self->{label} = $name;
	$self->{name} = $name;
	$self->{perpage} = $perpage;
	$self->parent->has_repeat;
}

sub name {
	my($self) = @_;
	$self->{name};
}

sub perpage {
	my($self) = @_;
	$self->{perpage};
}

sub has_repeat {
	my($self) = @_;
	$self->{grouped} = 1;
}

sub grouped {
	my($self) = @_;
	$self->{grouped};
}

sub innerfields {
	my($self) = @_;
	@{$self->{innerfields}};
}

sub add_innerfield {
	my($self, @fields) = @_;
	for my $field(@fields) {
		push @{$self->{innerfields}}, $field 
			unless grep {$_ eq $field} @{$self->{innerfields}};
	}
}

sub add_usefield {
	my($self, @fields) = @_;
	$self->parent->add_usefield(@fields) if $self->parent;
}

sub add_separator {
	my($self, $base, $shift) = @_;
	$base ||= 1;
	$shift ||= 0;
	push @{$self->{separators}}, {base => $base, shift => $shift};
}

sub add_separatorliteral {
	my($self, $literal) = @_;
	my $separators = $self->{separators};
	$separators->[$#$separators]->{literal} .= $literal;
}

sub separators {
	my($self) = @_;
	$self->{separators} ? @{$self->{separators}} : ();
}

sub make_report {
	my($self, $parent) = @_;
	my $report = Greese::ReportContainer::Repeat->new($parent, $self);
	for my $child($self->children) {
		$child->make_report($report);
	}
	$report;
}

#--------------------------------------------------------------------
package Greese::TemplateContainer::UseElse;
use vars qw(@ISA);
@ISA = qw(Greese::TemplateContainer);

sub add_child {
	my($self, $child) = @_;
	if( $self->{useelse} ) {
		push @{$self->{else}}, $child;
	} else {
		push @{$self->{children}}, $child;
	}
}

sub useelse {
	my($self, $useelse) = @_;
	$self->{useelse} = $useelse;
}

sub elsechildren {
	my($self) = @_;
	$self->{else} ? @{$self->{else}} : ();
}

#--------------------------------------------------------------------
package Greese::TemplateContainer::If;
use vars qw(@ISA);
@ISA = qw(Greese::TemplateContainer::UseElse);

sub init {
	my($self, $label, $condition) = @_;
	$self->{label} = $label;
	$self->{condition} = $condition;
	while( $condition =~ /%\[([^\(\)]+?)\]%/g ) {
		my $fname = $1;
		$self->parent->add_innerfield($fname);
		$self->parent->add_usefield($fname);
	}
}

sub label {
	my($self) = @_;
	$self->{label};
}

sub condition {
	my($self) = @_;
	$self->{condition};
}

sub make_report {
	my($self, $parent) = @_;
	my $report = Greese::ReportContainer::If->new($parent, $self);
	for my $child($self->children ) {
		$child->make_report($report);
	}
	$report->useelse(1);
	for my $child($self->elsechildren ) {
		$child->make_report($report);
	}
	$report;
}

#--------------------------------------------------------------------
package Greese::TemplateContainer::Select;
use vars qw(@ISA);
@ISA = qw(Greese::TemplateContainer::UseElse);

sub init {
	my($self, $label, $fromwhere) = @_;
	$self->{label} = $label;
	$self->{fromwhere} = $fromwhere;
	while( $fromwhere =~ /%\[([^\(\)]+?)\]%/g ) {
		my $fname = $1;
		$self->parent->add_innerfield($fname);
		$self->parent->add_usefield($fname);
	}
}

sub label {
	my($self) = @_;
	$self->{label};
}

sub fromwhere {
	my($self) = @_;
	$self->{fromwhere};
}

sub usefields {
	my($self) = @_;
	$self->{usefields} ? @{$self->{usefields}} : ();
}

sub add_usefield {
	my($self, @fields) = @_;
	for my $field(@fields) {
		push @{$self->{usefields}}, $field 
			unless grep {$_ eq $field} @{$self->{usefields}};
	}
}

sub add_innerfield {
	my($self, @fields) = @_;
	# do nothing
}

sub elseliteral {
	my($self) = @_;
	$self->{elseliteral};
}

sub add_elseliteral {
	my($self, $literal) = @_;
	$self->{elseliteral} = $literal;
}

sub alias {
	my($self, $field) = @_;
	if( defined $field ) {
		$self->{alias}{$field};
	} else {
		$self->{alias} || {};
	}
}

sub add_alias {
	my($self, $alias) = @_;
	if( $alias =~ /\s+as\s+(\S+)/i ) {
		my $asname = $1;
		$asname =~ s/^\"//;
		$asname =~ s/\"$//;
		$self->{alias}{$asname} = $alias;
	} else {
		die "alias() must include 'as'\n";
	}
}

sub perpage {
	my($self, $perpage) = @_;
	$self->{perpage} = $perpage if defined $perpage;
	$self->{perpage};
}

sub perpagename {
	my($self, $name) = @_;
	$self->{perpagename} = $name if defined $name;
	$self->{perpagename};
}

sub has_repeat {
	my($self) = @_;
	# do nothing
}

sub make_report {
	my($self, $parent) = @_;
	my $report = Greese::ReportContainer::Select->new($parent, $self);
	for my $child($self->children ) {
		$child->make_report($report);
	}
	$report->useelse(1);
	for my $child($self->elsechildren ) {
		$child->make_report($report);
	}
	$report;
}

#--------------------------------------------------------------------
package Greese::TemplateContainer::File;
use vars qw(@ISA);
@ISA = qw(Greese::TemplateContainer);

sub init {
	my($self, $template, $file) = @_;
	$self->{template} = $template;
	$self->{file} = $file;
}

sub template {
	my($self) = @_;
	$self->{template};
}

sub file {
	my($self) = @_;
	$self->{file};
}

sub parse {
	my($self) = @_;
	my $file = $self->{file};
	my $str;
	if( ref($file) eq 'SCALAR' ) {
		$str = $$file;
	} else {
		my $handle = new FileHandle $file;
		die "cannot open $file" unless $handle;
		local($/) = undef;
		$str = <$handle>;
	}
	$self->parse_str($str);
	# $self->detach_subfile($self->template);
}

sub LM_NORMAL { 1 }
sub LM_SEPARATOR { 2 }
sub LM_ELSE { 3 }

sub parse_str {
	my($self, $str) = @_;
	my $cur = $self;
	my @curselect;
	my $escchr = $self->template->esc_char;
	my $starttag = $self->template->start_tag;
	my $endtag = $self->template->end_tag;
	my $cmdptn = commandpattern($escchr, $starttag, $endtag);
	my $tmpcmdptn;
	my $literalmode = 0;
	while( $literalmode ? $str =~ /$tmpcmdptn/ : $str =~ /$cmdptn/ ) {
		my $literal = $`;
		my($label, $command) = ($1, $2);
		$str = $';
		if( $literal ne "" ) {
			$literal =~ s/\Q$escchr$starttag\E/$starttag/g 
				if !$literalmode && $escchr ne "";
			if( $literalmode == LM_SEPARATOR ) { 
				$cur->add_separatorliteral($literal);
			} elsif( $literalmode == LM_ELSE ) { 
				$cur->add_elseliteral($literal);
			} else {
				if( $cur->isa('Greese::TemplateItem::Literal') ) {
					$cur->add_string($literal);
				} else {
					Greese::TemplateItem::Literal->new($cur, $label, $literal);
				}
			}
		}
		$literalmode = 0;
		$command =~ s/^\s+//;
		$command =~ s/\s+$//;
		if( $command ne "" ) {
			if( $command =~ /^select\((.*)\)/ ) {
				$cur = Greese::TemplateContainer::Select->new($cur, $label, $1);
				unshift @curselect, $cur;
			} elsif( $command =~ /^tag\((.*)\)/ ) {
				my @spec = csv2array($1);
				if( @spec == 2 ) {
					($starttag, $endtag) = @spec;
				} elsif( @spec == 3 ) {
					($starttag, $endtag, $escchr) = @spec;
				} else {
					die "tag() must have 2 or 3 arguments";
				}
				$cmdptn = commandpattern($escchr, $starttag, $endtag);
			} elsif( $command =~ /^if\((.*)\)/ ) {
				$cur = Greese::TemplateContainer::If->new($cur, $label, $1);
			} elsif( $command =~ /^literal\((.*)\)/ ) {
				$cur = Greese::TemplateItem::Literal->new($cur, $label);
				#my $name = $1;
				#$tmpcmdptn = qr/\Q$starttag\Eend\(\Q$name\E\)\Q$endtag\E/;
				$tmpcmdptn = 
					commandpattern($escchr, $starttag, $endtag, $label, 'end');
				$literalmode = LM_NORMAL;
			} elsif( $command =~ /^comment\((.*)\)/ ) {
				# do nothing
			} elsif( $command =~ /^alias\((.*)\)/ ) {
				$cur->add_alias($1);
			} elsif( $command =~ /^repeat\((.*)\)/ ) {
				my $perpage = $1;
				die "repeat() argument must be an integer number\n"
					unless $perpage =~ /^\s*\d*\s*$/;
				my $name = $label;
				if( $perpage && $name eq '' ) {
					$name = "_" . $perpage . ":";
				}
				die "repeat($name) already exists\n" 
					if $label ne '' && exists $curselect[0]->{rnhash}{$name};
				die "at most one perpage specification for a file\n"
					if $perpage && $curselect[0]->perpage;
				$cur = Greese::TemplateContainer::Repeat->new($cur, $name, 
					$perpage);
				$curselect[0]->{rnhash}{$name} = $cur if $name ne '';
				$curselect[0]->perpagename($name) if $perpage;
				$curselect[0]->perpage($perpage) if $perpage;
			} elsif( $command =~ /^separator\((.*)\)/ ) {
				chklabel($cur, $label, $command);
				die "separator() must be between repeat() and end()\n"
					unless UNIVERSAL::isa($cur, 
						'Greese::TemplateContainer::Repeat');
				$cur->add_separator(csv2array($1));
				#$tmpcmdptn = qr/\Q$starttag\E(((\w+:)?(separator|end))\(.*?\))\Q$endtag\E/;
				$tmpcmdptn = commandpattern($escchr, $starttag, $endtag, 
					undef, 'separator', 'end');
				$literalmode = LM_SEPARATOR;
			} elsif( $command =~ /^else\(\)/ ) {
				chklabel($cur, $label, $command);
				die "else() must be between select() or if() and end()\n"
					unless UNIVERSAL::isa($cur, 
						'Greese::TemplateContainer::Select') ||
						UNIVERSAL::isa($cur, 
						'Greese::TemplateContainer::If');
				#$tmpcmdptn = 
				#	commandpattern($escchr, $starttag, $endtag, undef, 'end');
				#$literalmode = LM_ELSE;
				$cur->useelse(1);
			} elsif( $command =~ /^end\(\)/ ) {
				chklabel($cur, $label, $command);
				if( $cur->isa('Greese::TemplateContainer::Select') ) {
					shift @curselect;
				}
				$cur = $cur->parent;
				die "too much end()\n" unless $cur;
			} elsif( $command =~ /^embed\((.+)\)/ ) {
				my($file, @args) = csv2array($1);
				if( $file =~ /^<<(\w+)/ ) {
					my $end = $1;
					if( $str =~ /^\Q$end\E$/m ) {
						my $embed = $`;
						$str = $';
						$file = \$embed;
					} else {
						die "missing end mark '$end'\n";
					}
				} else {
					if( $file !~ /^\// ) {
						my $dir = $self->file;
						$dir =~ s/[^\\\/]+$//;
						$file = "$dir$file";
					}
					die "missing template file to embed: $file\n" 
						unless -f $file;
				}
				#die "self-embedding inhibited\n" if $file eq $self->file;
				#my $subtemplate = $self->template->new($file);
				#die "cannot make sub template\n" unless $subtemplate;
				my $subtemplate = $file;
				unshift @args, $subtemplate;
				Greese::TemplateItem::Embed->new($cur, \@args);
			} elsif( $command =~ /\(/ ) {
				Greese::TemplateItem::Expression->new($cur, $label, $command);
			} else {
				Greese::TemplateItem::Field->new($cur, $label, $command);
			}
		}
	}
	if( $str ne "" ) {
		die "missing $cur->{label}end() at the end of file\n"
			unless $cur->isa('Greese::TemplateContainer');
		$str =~ s/\Q$escchr$starttag\E/$starttag/g if $escchr ne "";
		Greese::TemplateItem::Literal->new($cur, undef, $str);
	}
	die "insufficient end() at end of file\n"
		unless UNIVERSAL::isa($cur, 'Greese::TemplateContainer::File');
}

sub make_report {
	my($self, $parent, $reportobj) = @_;
	my $report = Greese::ReportContainer::File->new($parent, $reportobj, $self);
	for my $child($self->children) {
		$child->make_report($report);
	}
	$report;
}

# NOT method
sub chklabel {
	my($cur, $label, $command) = @_;
	die "'$label$command' label mismatch (may be '$cur->{label}')\n" 
		if $label ne '' && $label ne $cur->{label};
}

sub commandpattern {
	my($escchr, $starttag, $endtag, $label, @commands) = @_;
	my $labelpart = $label ne "" ? qr/(\Q$label\E)/ : qr/([,\w]+:)?/;
	my $commands = join("|", map {"\Q$_\E"} @commands);
	my $cmdpart = @commands ? qr/(($commands)\(.*?\))/ : qr/(.+?)/;
	$escchr ne "" ? 
		qr/(?<!\Q$escchr\E)\Q$starttag\E$labelpart$cmdpart\Q$endtag\E/ :
		qr/\Q$starttag\E$labelpart$cmdpart\Q$endtag\E/;
}

sub csv2array {
	local($_) = @_;
	my @result;
	while( /\s*(\"(\"\"|.)*?\"|[^,]*)\s*(,|$)/g ) { 
		my($item, $punc) = ($1, $3);
		if( $item =~ /^\"/ ) {
			$item =~ s/^\"//;
			$item =~ s/\"$//;
			$item =~ s/\"\"/\"/g;
		}
		push(@result,$item); 
		last unless $punc eq ',';
	}
	@result;
}

#--------------------------------------------------------------------
# report classes
#--------------------------------------------------------------------
package Greese::Report;
use FileHandle;

sub new {
	my($class, $template) = @_;
	my $self = bless {template => $template}, $class;
	$self->{filenode} = $template->filenode->make_report(undef, $self);
	$self;
}

sub template {
	my($self) = @_;
	$self->{template};
}

sub filenode {
	my($self) = @_;
	$self->{filenode};
}

sub print {
	my($self, %param) = @_;
	$self->filenode->print(%param);
}

sub output {
	my($self, %param) = @_;
	$self->filenode->output(%param);
}

sub dump {
	my($self, $indent) = @_;
	$self->filenode->dump($indent);
}

# NOT method
sub fieldsvalue {
	my($tuple, @fields) = @_;
	$tuple ? map {$tuple->{$_}} @fields : map {undef} @fields;
}

sub make_filename {
	my($templatefile, $num) = @_;
	my $filename = $templatefile;
	$filename =~ s/\.[^\.]+$/.$num$&/ or $filename =~ s/$/.$num/;
	$filename;
}

sub quotename {
	my($name) = @_;
	$name =~ s/[^\s\.]+/\"$&\"/g unless $Greese::NOQUOTENAME;
	$name;
}

sub relpath {
	my($srcpath, $dstpath) = @_;
	return $dstpath if $dstpath =~ /^(https?:)?\//i;
	my @src = split('/', $srcpath);
	my @dst = split('/', $dstpath);
	while( @src && @dst ) {
		if( $src[0] eq $dst[0] ) {
			shift @src;
			shift @dst;
		} else {
			last;
		}
	}
	@src ? '../' x (scalar(@src) - 1) . join('/', @dst) :
		@dst ? join('/', @dst) : 
		$dstpath;
}

#--------------------------------------------------------------------
package Greese::ReportItem;
use vars qw(@ISA);
@ISA = qw(Greese::Item);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
}

sub template {
	my($self) = @_;
	$self->{template};
}

sub add_tuple {
	my($self, $tuple) = @_;
	$self->{value} = $self->template->value;
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{value};
}

sub select {
	my($self) = @_;
	my $result;
	my $parent = $self->parent;
	while( $parent ) {
		if( $parent->isa('Greese::ReportContainer::Select') ) {
			$result = $parent;
			last;
		}
		$parent = $parent->parent;
	}
	$result;
}

sub file {
	my($self) = @_;
	my $result;
	my $parent = $self->parent;
	while( $parent ) {
		if( $parent->isa('Greese::ReportContainer::File') ) {
			$result = $parent;
			last;
		}
		$parent = $parent->parent;
	}
	$result;
}

sub report {
	my($self) = @_;
	$self->file->report;
}

sub output {
	my($self, $results) = @_;
	push @$results, $self->value;
}


#--------------------------------------------------------------------
package Greese::ReportItem::Literal;
use vars qw(@ISA);
@ISA = qw(Greese::ReportItem);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
	$self->{value} = $self->template->value;
}

#--------------------------------------------------------------------
package Greese::ReportItem::Field;
use vars qw(@ISA);
@ISA = qw(Greese::ReportItem);

sub add_tuple {
	my($self, $tuple) = @_;
	if( $self->template->value =~ /^\'(.*)\'$/ ) {
		$self->{value} = $1;
	} elsif( $tuple ) {
		$self->{value} = $tuple->{$self->template->value};
	} else {
		$self->{value} = undef;
	}
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{value};
}

sub output {
	my($self, $results) = @_;
	my $tobj = $self->report->template;
	my $value;
	my $filter = $self->template->{label};
	if( $filter ) {
		$filter =~ s/:$//;
		$value = $self->value;
		my @filters = split /,/, $filter;
		if( $filters[0] ne 'raw' ) {
			$value = $tobj->escape($value);
		}
		for(@filters) {
			next if $_ eq 'raw';
			my $func = "filter_$_";
			$value = $tobj->$func($value) if $tobj->can($func);
		}
	} else {
		$value = $tobj->escape($self->value);
	}
	push @$results, $value;
}

#--------------------------------------------------------------------
package Greese::ReportItem::Expression;
use vars qw(@ISA);
@ISA = qw(Greese::ReportItem);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
	my $exp = $self->template->value;
	if( $exp =~ /^r?eval\((.+)\)/ ) {
		$self->{value} = $1;
	}
}

sub add_tuple {
	my($self, $tuple) = @_;
	my $exp = $self->template->value;
	if( $exp =~ /^sum\((.+)\)/ && $tuple ) {
		$self->{value} += $tuple->{$1};
	}
	if( $exp =~ /^max\((.+)\)/ && $tuple ) {
		$self->{value} = $tuple->{$1} if $self->{value} < $tuple->{$1};
	}
	if( $exp =~ /^min\((.+)\)/ && $tuple ) {
		$self->{value} = $tuple->{$1} if $self->{value} > $tuple->{$1};
	}
	if( $exp =~ /^r?eval\((.+)\)/ && $tuple ) {
		my $eval = $1;
		$eval =~ s/%\[([^\(\)]+?)\]%/$tuple->{$1}/ge;
		$self->{value} = $eval;
	}
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{value};
}

sub output {
	my($self, $results, %param) = @_;
	my $exp = $self->template->value;
	my $value;
	if( $exp =~ /^(sum|max|min)\((.+)\)/ ) {
		$value = $self->value;
	} elsif( $exp =~ /^(r?)eval\((.+)\)/ ) {
		my $r = $1;
		my $code = $self->value;
		$code =~ s/%\[(\w+\(.*?\))\]%/$self->expvalue($1, \%param)/ge;
		my $result = $Safe->reval($code);
		die "error '$@' in eval($exp)" if $@;
		$value = $r ? $result : '';
	} else {
		$value = $self->expvalue($exp, \%param);
	}
	my $tobj = $self->report->template;
	my $filter = $self->template->{label};
	if( $filter ) {
		$filter =~ s/:$//;
		my @filters = split /,/, $filter;
		if( $filters[0] ne 'raw' ) {
			$value = $tobj->escape($value);
		}
		for(@filters) {
			next if $_ eq 'raw';
			my $func = "filter_$_";
			$value = $tobj->$func($value) if $tobj->can($func);
		}
	} else {
		$value = $tobj->escape($value);
	}
	push @$results, $value;
}

#--------------------------------------------------------------------
package Greese::ReportItem::Embed;
use vars qw(@ISA);
@ISA = qw(Greese::ReportItem);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
	my %args = %{$self->template->{args}};
	$self->{args} = \%args;
}

sub add_tuple {
	my($self, $tuple, %param) = @_;
	$self->{args} = {};
	for my $key(keys %{$self->template->{args}}) {
		my $value = $self->template->{args}{$key};
		$value =~ s/%\[([^\(\)]+?)\]%/$tuple->{$1}/ge;
		# $value =~ s/%\[(param\((.+?)\)|(.+?))\]%/$2 ? $param{$2} : $tuple->{$3}/ge;
		$self->{args}{$key} = $value;
	}
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{count};
	my %args = %{$self->template->{args}};
	$self->{args} = \%args;
}

sub output {
	my($self, $results, %param) = @_;
	for my $key(keys %{$self->{args}}) {
		$self->{args}{$key} =~ 
			s/%\[(\w+\(.*?\))\]%/$self->expvalue($1, \%param)/ge;
	}
	my $subtemplate = $self->report->template->new(
		$self->template->{subtemplate});
	my $subrep = Greese::Report->new($subtemplate);
	push @$results, $subrep->output(dbihandle => $param{dbihandle}, 
		%{$self->{args}});
}

#--------------------------------------------------------------------
package Greese::ReportContainer;
use vars qw(@ISA);
@ISA = qw(Greese::Container);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
}

sub template {
	my($self) = @_;
	$self->{template};
}

sub inc_count {
	my($self, $name) = @_;
	$self->{count}{$name}++;
	$self->parent->inc_count($name) if $self->parent;
}

sub count {
	my($self, $name) = @_;
	$self->{count}{$name} + 0;
}

sub pcount {
	my($self, $name) = @_;
	for( my $obj = $self; $obj; $obj = $obj->parent ) {
		return $obj->{count}{$name} + 0 if exists $obj->{count}{$name};
	}
	0;
}

sub ptuples {
	my($self) = @_;
	for( my $obj = $self; $obj; $obj = $obj->parent ) {
		return $obj->{tuples} if exists $obj->{tuples};
	}
	0;
}

#--------------------------------------------------------------------
package Greese::ReportContainer::RepeatInstance;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer);

sub init {
	my($self, $template, $count, @children) = @_;
	$self->{template} = $template;
	$self->{children} = \@children;
	my $name = $template->name;
	$self->{count}{$name} = $count if $name ne '';
	for my $child(@children) {
		$child->set_parent($self);
	}
}

sub add_tuple {
	my($self, $tuple, %param) = @_;
	$self->{tuples}++ if $tuple;
	for my $child($self->children) {
		$child->add_tuple($tuple, %param);
	}
}

sub tuples {
	my($self) = @_;
	$self->{tuples};
}

sub output {
	my($self, $results, %param) = @_;
	for my $child($self->children) {
		$child->output($results, %param);
	}
}

#--------------------------------------------------------------------
package Greese::ReportContainer::Repeat;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer);

sub make_instance {
	my($self) = @_;
	my $name = $self->template->name;
	my $count;
	if( $name ne '' ) {
		$self->inc_count($name);
		$count = $self->count($name);
	}
	my $instance = Greese::ReportContainer::RepeatInstance->new_instance($self, 
		$self->template, $count, map {$_->clone} $self->children);
	$instance;
}

sub add_tuple {
	my($self, $tuple, %param) = @_;
	return unless $tuple;
	my $instance;
	if( $self->template->grouped ) {
		my @values = Greese::Report::fieldsvalue($tuple, 
			$self->template->innerfields);
		my $key = join($;, @values);
		$instance = $self->{grouphash}{$key};
		unless( $instance ) {
			$instance = $self->make_instance;
			$self->{grouphash}{$key} = $instance;
		}
	} else {
		$instance = $self->make_instance;
	}
	$instance->add_tuple($tuple, %param);
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{count};
	$self->clr_instance;
}

sub separator {
	my($self, $count) = @_;
	$count++;
	my $result;
	for my $separator( $self->template->separators ) {
		my $base = $separator->{base};
		my $shift = $separator->{shift};
		my $literal = $separator->{literal};
		if( $count % $base == $shift ) {
			$result .= $literal;
		}
	}
	$result;
}

sub output {
	my($self, $results, %param) = @_;
	my $perpage = $self->template->perpage;
	my($count, $countend);
	if( $param{_page} ) {
		if( $perpage ) {
			my $start = $param{_start};
			my $end = $start + $param{_count} - 1;
			delete $param{_page};
			delete $param{_start};
			delete $param{_count};
			$count = $start;
			$countend = $end;
			for my $instance(($self->instances)[$start..$end]) {
				$instance->output($results, %param);
				push @$results, $self->separator($count) if $count < $countend;
				$count++;
			}
		} else {
			my @instances = $self->instances;
			$count = 0;
			$countend = $#instances;
			for my $instance(@instances) {
				my $partcount = $instance->count($param{perpagename});
				if( $partcount <= $param{_start} ) {
					$param{_start} -= $partcount;
				} else {
					my $leftcount = 
						$param{_start} + $param{_count} > $partcount ? 
						$param{_start} + $param{_count} - $partcount : 
						0;
					$param{_count} -= $leftcount;
					$instance->output($results, %param);
					last unless $leftcount;
					push @$results, $self->separator($count) 
						if $count < $countend;
					$param{_start} = 0;
					$param{_count} = $leftcount;
				}
				$count++;
			}
		}
	} else {
		my @instances = $self->instances;
		$count = 0;
		$countend = $#instances;
		for my $instance(@instances) {
			$instance->output($results, %param);
			push @$results, $self->separator($count) if $count < $countend;
			$count++;
		}
	}
}

#--------------------------------------------------------------------
package Greese::ReportContainer::UseElse;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer);

sub add_child {
	my($self, $child) = @_;
	if( $self->{useelse} ) {
		push @{$self->{else}}, $child;
	} else {
		push @{$self->{children}}, $child;
	}
}

sub useelse {
	my($self, $useelse) = @_;
	$self->{useelse} = $useelse;
}

sub elsechildren {
	my($self) = @_;
	$self->{else} ? @{$self->{else}} : ();
}

#--------------------------------------------------------------------
package Greese::ReportContainer::If;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer::UseElse);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
	$self->{condition} = $template->condition;
}

sub add_tuple {
	my($self, $tuple, %param) = @_;
	my $condition = $self->template->condition;
	# $condition =~ s/%\[(param\((.+?)\)|(.+?))\]%/$2 ? $param{$2} : $tuple->{$3}/ge;
	$condition =~ s/%\[([^\(\)]+?)\]%/$tuple->{$1}/ge;
	$self->{condition} = $condition;
	for my $child($self->children, $self->elsechildren) {
		$child->add_tuple($tuple, %param);
	}
}

sub clr_tuple {
	my($self) = @_;
	delete $self->{count};
	for my $child($self->children, $self->elsechildren) {
		$child->clr_tuple;
	}
}

sub output {
	my($self, $result, %param) = @_;
	my $condition = $self->{condition};
	# $condition =~ s/%\[param\((.+?)\)\]%/$param{$1}/ge;
	$condition =~ s/%\[(\w+\(.*?\))\]%/$self->expvalue($1, \%param)/ge;
	my $evaled = $Safe->reval($condition);
	die "error '$@' in eval($condition)" if $@;
	if( $evaled ) {
		for my $child($self->children) {
			$child->output($result, %param);
		}
	} else {
		for my $child($self->elsechildren) {
			$child->output($result, %param);
		}
	}
}

sub count {
	my($self, $name) = @_;
	$self->pcount($name);
}

sub tuples {
	my($self) = @_;
	$self->ptuples;
}

#--------------------------------------------------------------------
package Greese::ReportContainer::Select;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer::UseElse);

sub init {
	my($self, $template) = @_;
	$self->{template} = $template;
	$self->fromwhere($self->template->fromwhere);
}

sub report {
	my($self) = @_;
	$self->file->report;
}

sub add_where {
	my($self, @conditions) = @_;
	$self->fromwhere(_add_where($self->fromwhere, @conditions));
}

# NOT method
sub _add_where {
	my($fromwhere, @conditions) = @_;
	my $result = $fromwhere;
	$result =~ s/\s/ /g;
	for my $condition(@conditions) {
		next if $condition eq '';
		$result =~ s/ where / where $condition and /i or
		$result =~ s/( order | group |$)/ where $condition$1/i;
	}
	$result;
}

sub fromwhere {
	my($self, $fromwhere) = @_;
	$self->{fromwhere} = $fromwhere if defined $fromwhere;
	$self->{fromwhere};
}

sub fromvar {
	my($self, $label, $param) = @_;
	my $var;
	if( $self->{fromwhere} =~ /^\s*(var|param)\((.+)\)\s*$/ ) {
		my $pname = $2;
		$var = $param->{"$label$pname"} 
			or die "missing parameter '$label$pname'";
		die "parameter '$pname' must be an array ref" 
			unless ref($var) eq 'ARRAY';
	} elsif( $self->{fromwhere} =~ /^\s*eval\((.+)\)\s*$/ ) {
		my $evalstr = $1;
		my @var = $Safe->reval($evalstr);
		die "error '$@' in eval($evalstr)" if $@;
		$var = \@var;
	}
	$var;
}

sub tuples {
	my($self) = @_;
	$self->{tuples};
}

sub pages {
	my($self) = @_;
	my $perpage = $self->template->perpage;
	return unless $perpage;
	my $count = $self->count($self->template->perpagename);
	return 0 unless $count;
	int(($count - 1) / $perpage) + 1;
}

sub setparam {
	my($self, %param) = @_;
	my $fromwhere = $self->fromwhere;
	# $fromwhere =~ s/%\[param\((.+?)\)\]%/$param{$1}/ge;
	$fromwhere =~ s/%\[(\w+\(.*?\))\]%/$self->expvalue($1, \%param)/ge;
	$self->fromwhere($fromwhere);
}

sub add_tuple {
	my($self, $tuple, %param) = @_;
	my $fromwhere = $self->template->fromwhere;
	# $fromwhere =~ s/%\[(param\((.+?)\)|(.+?))\]%/$2 ? $param{$2} : $tuple->{$3}/ge;
	$fromwhere =~ s/%\[([^\(\)]+?)\]%/$tuple->{$1}/ge;
	$self->fromwhere($fromwhere);
}

sub _add_tuple {
	my($self, $tuple, %param) = @_;
	$self->{tuples}++ if $tuple;
	for my $child($self->children) {
		$child->add_tuple($tuple, %param);
	}
}

sub clr_tuple {
	my($self) = @_;
	$self->{tuples} = 0;
	delete $self->{count};
	for my $child($self->children) {
		$child->clr_tuple;
	}
}

sub read_data {
	my($self, %param) = @_;
	$self->clr_tuple;
	my $label = $self->template->label;
	$self->{param} = \%param;
	$self->setparam(%param);
	my $var = $self->fromvar($label, \%param);
	if( ref($var) eq 'ARRAY' ) {
		if( !@$var ) {
			$self->{empty} = 1;
			$self->_add_tuple(undef, %param) if $param{"${label}empty"};
		} else {
			if( ref($var->[0]) eq 'HASH' && ref($var->[1]) ne 'HASH' ) {
				my $varparam = shift @$var;
				%param = (%param, %$varparam);
			}
			if( ref($var->[0]) eq 'ARRAY' ) {
				if( @$var <= 1 ) {
					$self->{empty} = 1;
					$self->_add_tuple(undef, %param) if $param{"${label}empty"};
				} else {
					my $idx = 1;
					my %fieldhash = map {($_, $idx++)} @{$var->[0]};
					for( my $j = 1; $j < @$var; $j++ ) {
						$self->_add_tuple([\%fieldhash, @{$var->[$j]}], %param);
					}
				}
			} elsif( ref($var->[0]) eq 'HASH' ) {
				for my $hash(@$var) {
					$self->_add_tuple($hash, %param);
				}
			} else {
				if( @$var <= 1 ) {
					$self->{empty} = 1;
					$self->_add_tuple(undef, %param) if $param{"${label}empty"};
				} else {
					for( my $j = 1; $j < @$var; $j++ ) {
						$self->_add_tuple({$var->[0], $var->[$j]}, %param);
					}
				}
			}
		}
	} else {
		die "missing dbihandle parameter\n" unless $param{dbihandle};
		my $dbihandle = $param{dbihandle};
		$self->{dbihandle} = $dbihandle;
		my @fields = $self->template->usefields;
		my $fromwhere = $self->add_where(
			$param{"_${label}where_used"} ? '' : $param{"${label}where"});
		$param{"_${label}where_used"} = 1;
		$param{fromwhere} = $fromwhere;
		if( @fields && !$fromwhere ) {
			die "missing FROM clause\n";
		} else {
			my $idx = 1;
			my %fieldhash = map {($_, $idx++)} @fields;
			my $cols = join(",", map {
				$self->template->alias($_) ? 
					$self->template->alias($_) :
					Greese::Report::quotename($_)
				} @fields);
			$cols ||= $Greese::DEFAULTCOLUMN;
			my $sql = "select $cols $fromwhere";
			my $sth = $dbihandle->prepare($sql)
				or die $sql.':'.$dbihandle->errstr;
			$sth->execute
				or die $sql.':'.$sth->errstr;
			my(@row, $rows);
			while( @row = $sth->fetchrow_array ) {
				$rows++;
				$self->_add_tuple([\%fieldhash, @row], %param);
			}
			if( !$rows ) {
				$self->{empty} = 1;
				$self->_add_tuple(undef, %param) if $param{"${label}empty"};
			}
		}
	}
	return %param;
}

sub setpageparam {
	my($self, $param) = @_;
	my $label = $self->template->label;
	my $perpage = $self->template->perpage;
	if( $param->{"${label}page"} ) {
		die "page specification for no perpage\n" unless $perpage;
		$param->{_page} = $param->{"${label}page"};
	} else {
		delete $param->{_page};
	}
	$param->{_page} = 1 if $perpage && !$param->{_page};
	if( $param->{_page} ) {
		$param->{_pages} = $self->pages;
		$param->{perpagename} = $self->template->perpagename;
		my $page = $param->{_page};
		my $perpage = $self->template->perpage;
		my $total = $self->count($self->template->perpagename);
		my $start = ($page - 1) * $perpage;
		die "no page $page\n" if $start >= $total && !$param->{"${label}empty"};
		my $count = $start + $perpage >= $total ? $total - $start : $perpage;
		$param->{_start} = $start;
		$param->{_count} = $count;
	}
}

sub output {
	my($self, $result, %param) = @_;
	%param = $self->read_data(%param);
	#if( $self->{empty} && $self->template->elseliteral ) {
	#	push @$result, $self->template->elseliteral;
	if( $self->{empty} && $self->elsechildren ) {
		for my $child($self->elsechildren) {
			$child->output($result, %param);
		}
	} else {
		$self->setpageparam(\%param);
		for my $child($self->children) {
			$child->output($result, %param);
		}
	}
}

#--------------------------------------------------------------------
package Greese::ReportContainer::File;
use FileHandle;
use vars qw(@ISA);
@ISA = qw(Greese::ReportContainer);

sub init {
	my($self, $report, $template) = @_;
	$self->{report} = $report;
	$self->{template} = $template;
}

sub report {
	my($self) = @_;
	$self->{report};
}

sub print {
	my($self, %param) = @_;
	$self->_print($param{outfile} || '-', %param);
}

sub _print {
	my($self, $ofile, %param) = @_;
	my $handle = new FileHandle ">$ofile";
	die "cannot open $ofile\n" unless $handle;
	print "writing $ofile..." if $param{verbose};
	print $handle $self->output(%param);
	print "done\n" if $param{verbose};
}

sub output {
	my($self, %param) = @_;
	my @result;
	for my $child($self->children) {
		$child->output(\@result, %param);
	}
	$self->report->template->postfilter(join('', @result));
}

1;
__END__
=pod

=head1 NAME

Greese - Generic Report Engine by Extended Select Expression

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 AUTHOR

Sey Nakajima <nakajima@netstock.co.jp>

=head1 SEE ALSO

=cut
